"""
Copyright 2008  Blub

FileServerHandler, a class which implements a simple handler for HttpServer
that serves files which are inside a configured directory.
This file is part of PebCache.

PebCache is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation, version 2 of the License.

PebCache is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PebCache.  If not, see <http://www.gnu.org/licenses/>.
"""

##builtin
import logging
import mimetypes
import os

##own
from Http import HttpUtilities




class FileServerHandler:
    def __init__(self, handlerName, config, ownDest, path):
        self.config = config
        self.ownDest = ownDest
        self.path = path
        self.log = logging.getLogger(handlerName)
        self._checkConfig()   
        self.rootPath = self.config.get('RootFolder')
        if len(self.rootPath) == 0:
            self.rootPath = os.path.normpath(self.path)
        else:
            if self.rootPath[0]=='/':
                self.rootPath = os.path.normpath(self.rootPath)
            else:
                self.rootPath = os.path.normpath(os.path.join(self.path, self.rootPath))
        
        
    def _checkConfig(self):
        #set default values
        configDefaults = {'RootFolder':(u'Html', 'unicode'),
                          'DefaultFile':(u'index.html', 'unicode')}
        self.config.addDefaults(configDefaults)
        
        
    def doGet(self, connId, sourceDest, request):
        header = request['header']
        self.log.info('Conn %5i: Got Request "%s"', connId, HttpUtilities.joinUrl(header['url']))
        response = {}
        
        #handle url
        requestedUrl = header['url']['path']
        if len(requestedUrl) == 0:
            requestedUrl = '/'
            
        requestedUrl = os.path.normpath(requestedUrl)
        if requestedUrl == '/':
            #base folder, add default file
            requestedUrl = os.path.normpath(os.path.join(requestedUrl, self.config.get('DefaultFile')))
        
        #remove first slash
        if requestedUrl[0] == '/':
            requestedUrl = requestedUrl[1:]
            
        #join with base folder
        requestedUrl = os.path.normpath(os.path.join(self.rootPath, requestedUrl))
        self.log.debug('Conn %5i: Normalised Request URL: "%s".',connId,  requestedUrl)
        
        #get common part of both base folder and request url
        #to check if the request url is inside the base folder
        common = os.path.commonprefix([self.rootPath, requestedUrl])
        self.log.debug('Conn %5i: common: "%s".', connId, common)
    
        if not common == self.rootPath:
            #below base folder, return File Not Found
            response['code'] = 404
        else:
            #inside base folder
            if not os.path.exists(requestedUrl):
                #requested file or directory doesn't exist
                response['code'] = 404
            else:
                if not os.path.isfile(requestedUrl):
                    #directory requests are not supported
                    response['code'] = 403
                else:
                    #file request
                    mime = mimetypes.guess_type(os.path.splitext(requestedUrl)[1], strict=False)[0]
                    if mime is None:
                        mime = 'application/binary'
                    response['code'] = 200
                    response['type'] = mime
                    response['file'] = requestedUrl
        return response
