"""
Copyright 2009  Blub

GcacheDatabase, a class which implements a database backend for GcacheHandler.
This file is part of PebCache.

PebCache is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation, version 2 of the License.

PebCache is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PebCache.  If not, see <http://www.gnu.org/licenses/>.
"""

##builtin
from time import time
import logging
import os
import threading
from sqlite3 import dbapi2 as sqlite




class GcacheDatabase:
    def __init__(self, handlerName, path, config):
        self.config = config
        self.dbPath = os.path.join(path, self.config.get('DatabaseFile'))
        self.log = logging.getLogger(handlerName+'-DB')        
        
        if os.path.isfile(self.dbPath)==0:
            self.log.warn('Database not found, creating one')
            self._createDatabase()
            
        self.db = sqlite.connect(self.dbPath)
        self.lock = threading.Lock()
        
        
    def _createDatabase(self):
        connection = sqlite.connect(self.dbPath)
        cursor = connection.cursor()
        cursor.execute('CREATE TABLE hosts ('\
                       'hostId INTEGER PRIMARY KEY,'\
                       'destination TEXT UNIQUE,'\
                       'addTime INTEGER'\
                       ');')
        cursor.execute('CREATE TABLE caches ('\
                       'cacheId INTEGER PRIMARY KEY,'\
                       'url TEXT UNIQUE,'\
                       'addTime INTEGER'\
                       ');')
        connection.commit()
        connection.close()
    
    
    ##HOSTS##
    
    def _haveHost(self, destination):
        cursor = self.db.cursor()
        cursor.execute('SELECT hostId FROM hosts WHERE destination=?', (destination,))
        result = cursor.fetchone()
        return (result is not None)
    
    
    def _getHostCount(self):
        cursor = self.db.cursor()
        cursor.execute('SELECT count() FROM hosts')
        return cursor.fetchone()[0]
    
    
    def _removeOldestHost(self):
        cursor = self.db.cursor()
        cursor.execute('SELECT hostId FROM hosts ORDER BY addTime ASC LIMIT 1')
        hostId = cursor.fetchone()[0]
        cursor.execute('DELETE FROM hosts WHERE hostId=?', (hostId,))
        self.db.commit()
    
    
    def _addHost(self, destination):
        if not self._haveHost(destination):
            if self._getHostCount() >= self.config.get('HostReturnNum'):
                self._removeOldestHost()
            cursor = self.db.cursor()
            cursor.execute('INSERT INTO hosts VALUES (null, ?, ?);', (destination, int(round(time(), 0))))
            self.db.commit()
        
        
    def _getHosts(self, max):
        cursor = self.db.cursor()
        if max==-1:
            cursor.execute('SELECT destination FROM hosts ORDER BY addTime DESC')
        else:
            cursor.execute('SELECT destination FROM hosts ORDER BY addTime DESC LIMIT ?', (max,))
        hosts = cursor.fetchall()
        return [host[0] for host in hosts]
    
    
    ##CACHES##
    
    def _haveCache(self, url):
        cursor = self.db.cursor()
        cursor.execute('SELECT cacheId FROM caches WHERE url=?', (url,))
        result = cursor.fetchone()
        return (result is not None)
    
    
    def _getCacheCount(self):
        cursor = self.db.cursor()
        cursor.execute('SELECT count() FROM caches')
        return cursor.fetchone()[0]
    
    
    def _removeOldestCache(self):
        cursor = self.db.cursor()
        cursor.execute('SELECT cacheId FROM caches ORDER BY addTime ASC LIMIT 1')
        hostId = cursor.fetchone()[0]
        cursor.execute('DELETE FROM caches WHERE cacheId=?', (hostId,))
        self.db.commit()
        
        
    def _addCache(self, url):
        if not self._haveCache(url):
            if self._getCacheCount() >= self.config.get('CacheReturnNum'):
                self._removeOldestCache()
            cursor = self.db.cursor()
            cursor.execute('INSERT INTO caches VALUES (null, ?, ?);', (url, int(round(time(), 0))))
            self.db.commit()
        
        
    def _getCaches(self, max):
        cursor = self.db.cursor()
        if max==-1:
            cursor.execute('SELECT url FROM caches ORDER BY addTime DESC')
        else:
            cursor.execute('SELECT url FROM caches ORDER BY addTime DESC LIMIT ?', (max,))
        caches = cursor.fetchall()
        return [cache[0] for cache in caches]    
    
    
    ##EXTERNAL
        
    def addHost(self, destination):
        self.lock.acquire()
        self._addHost(destination)
        self.lock.release()
        
        
    def getHosts(self, max=-1):
        self.lock.acquire()
        hosts = self._getHosts(max)
        self.lock.release()
        return hosts
        
        
    def addCache(self, url):
        self.lock.acquire()
        self._addCache(url)
        self.lock.release()
        
        
    def getCaches(self, max=-1):
        self.lock.acquire()
        caches = self._getCaches(max)
        self.lock.release()
        return caches        
        
        
    def cleanup(self):
        self.lock.acquire()
        self.db.close()    
        self.lock.release()