"""
Copyright 2009  Blub

GcacheHandler, a class which implements a gnutella webcache as a simple handler for HttpServer.
This file is part of PebCache.

PebCache is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation, version 2 of the License.

PebCache is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PebCache.  If not, see <http://www.gnu.org/licenses/>.
"""

##builtin
import logging

##own
from Http import HttpUtilities
from Gcache.GcacheDatabase import GcacheDatabase




class GcacheHandler:
    def __init__(self, handlerName, config, ownDest, path, version):
        self.config = config
        self.ownDest = ownDest
        self.path = path
        self.version = version
        self.log = logging.getLogger(handlerName)
        self._checkConfig()
        self.db = GcacheDatabase(handlerName, self.path, self.config)       
        
        
    def _checkConfig(self):
        #set default values
        configDefaults = {'DatabaseFile':(u'gcache.db', 'unicode'),
                          'HostReturnNum':(20, 'int'),
                          'CacheReturnNum':(20, 'int')}
        self.config.addDefaults(configDefaults)
    
    
    def _getNumberGetParas(self, paras):
        getParas = ('ping', 'statfile', 'hostfile', 'urlfile')
        count = 0
        for para in paras.iterkeys():
            if para in getParas:
                if paras[para]=='1':
                    count += 1
        return count
    
    
    def _getNumberPutParas(self, paras):
        putParas = ('ip', 'url')
        count = 0
        for para in paras.iterkeys():
            if para in putParas:
                count += 1
        return count
    
    
    def _checkHost(self, updateDest):
        if len(updateDest)==516 and updateDest[-4:]=='AAAA':
            return True
        else:
            return False
    
    
    def _checkCache(self, updateUrl):
        result = False
        if updateUrl[:11]=='http://i2p/' and \
          ((len(updateUrl[11:])==516 and updateUrl[-4:]=='AAAA') or (len(updateUrl[11:])==517 and updateUrl[-5:]=='AAAA/')):
            #i2p url with destination instead of hostname
            result = True
        else:
            firstSlash = updateUrl.find('/', 7)
            i2pSign = updateUrl.find('.i2p/', 7)
            if updateUrl[:7]=='http://' and firstSlash == i2pSign+4 and updateUrl.count('?')==0:
                #i2p url with hostname and without parameters
                result = True
        return result
    
    
    def doGet(self, connId, sourceDest, request):
        header = request['header']
        urlParameters = header['url'].get('parameter', {})
        self.log.info('Conn %5i: Got Request "%s"', connId, HttpUtilities.joinUrl(header['url']))
        
        #response structure
        response = {}
        
        #get number of get and put requests
        numPutRequests = self._getNumberPutParas(urlParameters)
        numGetRequests = self._getNumberGetParas(urlParameters)
        
        #check for wrong requests
        if numGetRequests > 0 and numPutRequests > 0:
            #put and get at the same time is not supported
            response['code'] = 200
            response['type'] = 'text/plain'
            response['string'] = 'ERROR: Bad formatted query, put and get requests may not be done in the same query.\r\n\r\n'
        
        elif numPutRequests > 0:
            #put request
            data = ['OK\r\n']
            if 'ip' in urlParameters:
                #ip update                
                if self._checkHost(urlParameters['ip']):
                    #good host
                    self.db.addHost(urlParameters['ip'])
                else:
                    data.append('WARNING: Bad Ip\r\n')
            
            if 'url' in urlParameters:
                #urlupdate
                if self._checkCache(urlParameters['url']):
                    #good Cache
                    self.db.addCache(urlParameters['url'])
                else:
                    data.append('WARNING: Bad Url\r\n')
                
            data.append('\r\n')
            response['code'] = 200
            response['type'] = 'text/plain'
            response['string'] = ''.join(data)
            
        elif numGetRequests > 1:
            #too many get requests
            response['code'] = 200
            response['type'] = 'text/plain'
            response['string'] = 'ERROR: Bad formatted query, only one get request may be done in one query.\r\n\r\n'
            
        elif numGetRequests == 1:
            #get request
            if 'ping' in urlParameters:
                response['code'] = 200
                response['type'] = 'text/plain'
                response['string'] = 'PONG PebCache '+self.version+'\r\n\r\n'
                
            elif 'statfile' in urlParameters:
                response['code'] = 200
                response['type'] = 'text/plain'
                response['string'] = '0\r\n0\r\n0\r\n'
                
            elif 'hostfile' in urlParameters:
                hosts = self.db.getHosts()
                hosts.append('')
                response['code'] = 200
                response['type'] = 'text/plain'
                response['string'] = '\r\n'.join(hosts)
                
            elif 'urlfile' in urlParameters:
                caches = self.db.getCaches()
                caches.append('')
                response['code'] = 200
                response['type'] = 'text/plain'
                response['string'] = '\r\n'.join(caches)
                
        else:
            #whatever
            response['code'] = 200
            response['type'] = 'text/plain'
            response['string'] = 'ERROR: Bad formatted query: unknown parameters or none.\r\n\r\n'
    
        return response
    
    
    def cleanup(self):
        self.db.cleanup()
