"""
Copyright 2009  Blub

HttpRequestParser, a general class which can be used to parse standard http requests.
This file is part of PebCache.

PebCache is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation, version 2 of the License.

PebCache is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PebCache.  If not, see <http://www.gnu.org/licenses/>.
"""

##builtin
from collections import deque
import logging

##own
from Conversion import hexToInt
from Http import HttpUtilities
from Http.HttpUtilities import HttpUrlParseException




class HttpRequestParserException(Exception):
    def __init__(self, reason, *args):
        self.reason = reason % args
        Exception.__init__(self, self.reason)
        
    def getReason(self):
        return self.reason
        



class HttpRequestParser:
    def __init__(self, connId, maxHeaderSize=1024, maxDataSize=1048576):
        self.connId = connId
        self.userHeaderSizeLimit = maxHeaderSize
        self.userDataSizeLimit = maxDataSize
        
        #data
        self.header = None
        self.data = None
        self.dataSize = 0
        self.maxDataSize = 0
        
        #state
        self.step = 'header'
        self.newline = None
        self.transferEncoding = None
        self.finished = False
        
        #buffers
        self.buffer = deque()
        self.bufferSize = 0
        self.maxBufferSize = self.userHeaderSizeLimit
        
        #log
        self.log = logging.getLogger('HttpRequestParser')
        
    
    ##internal functions - buffer
    
    def _setMaxBufferSize(self, bytes):
        self.log.debug('Conn %5i: Setting max buffer size to "%i" bytes', self.connId, bytes)
        self.maxBufferSize = bytes
    
    
    def _getAllChunks(self, clear=True):
        data = ''.join(self.buffer)
        self.buffer.clear()
        self.bufferSize = 0
        return data
    
        
    def _getLastChunk(self):
        if len(self.buffer) == 0:
            lastChunk = ''
        else:
            lastChunk = self.buffer.pop()
            self.bufferSize -= len(lastChunk)
        return lastChunk
    
    
    def _storeChunk(self, data):
        #store data in buffer
        self.buffer.append(data)
        self.bufferSize += len(data)
        
        #check size
        if self.bufferSize > self.maxBufferSize:
            raise HttpRequestParserException('buffer size exceeds limit while receiving "%s"', self.step)
      
    
    def _clearBuffer(self):
        self.buffer.clear()
        self.bufferSize = 0
        
        
    ##internal functions - data buffer
    
    def _increaseDataSizeLimit(self, additionalBytes):
        if self.maxDataSize + additionalBytes > self.userDataSizeLimit:
            raise HttpRequestParserException('data size exceeds limit while receiving "%s"', self.step)
        else:
            self.maxDataSize += additionalBytes
            self.log.debug('Conn %5i: Increasing max data size by "%i" bytes to "%i" bytes', self.connId, additionalBytes, self.maxDataSize)
            
    
    def _storeData(self, data):
        #store data
        self.data.append(data)
        self.dataSize += len(data)
        
        if self.dataSize > self.maxDataSize:
            raise HttpRequestParserException('data size exceeds exptected size while receiving "%s"', self.step)
        
    
    def _getData(self):
        if self.data is None:
            data = None
        else:
            data = ''.join(self.data)
        return data
    
        
    ##internal functions - header
    
    def _parseHeaderValues(self, rawHeader, headerName='header'):
        headerValues = {}
        
        #first merge everything
        place = 0
        while place < len(rawHeader):
            headerLine = rawHeader[place]
            if ':' in headerLine:
                #go on
                place += 1
            else:
                #merge
                if place == 0:
                    raise HttpRequestParserException('First key:value line in %s is invalid: "%s"', headerName, headerLine)
                else:
                    rawHeader[place-1] += headerLine.lstrip()
                    del rawHeader[place]
                    
        #now parse it
        for headerLine in rawHeader:
            #one header line
            headerLine = headerLine.split(':')
            key = headerLine[0].lower()
            value = ':'.join(headerLine[1:]).lstrip()
            if ',' in value:
                #multiple values
                value = list([part.strip() for part in value.split(',') if not part.strip() == ''])
                if len(value) == 0:
                    value = ''
                
            if len(key) == 0:
                raise HttpRequestParserException('Invalid key:value line in %s, key is empty: "%s"', headerName, headerline)
            else:
                headerValues[key] = value

        return headerValues
        
    
    def _parseHeader(self, header):
        header = header.split(self.newline)
        headerStatus = [part for part in header[0].split(' ') if not part == '']
        headerValues = self._parseHeaderValues(header[1:])
        
        #check for invalid status lines
        if not len(headerStatus) == 3:
            raise HttpRequestParserException('Invalid header status line: "%s"', header[0])
        
        if not headerStatus[2].upper() in ('HTTP/1.0', 'HTTP/1.1'):
            raise HttpRequestParserException('Unsupported http version: "%s"', headerStatus[2])
        
        #check type of response
        if not headerStatus[0].upper() in ('OPTIONS', 'GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'TRACE', 'CONNECT'):
            raise HttpRequestParserException('Unsupported http method: "%s"', headerStatus[0])
        
        #parsee url
        try:
            url = HttpUtilities.splitUrl(headerStatus[1], allowRelative=True)
        except HttpUrlParseException, e:
            raise HttpRequestParserException('Failed to parse url: %s', e.reason)
        
        #remember header
        self.header = {'method':headerStatus[0].upper(),
                       'url':url,
                       'version':headerStatus[2].upper(),
                       'values':headerValues}
                    
        #determine transfer length and encoding
        if 'transfer-encoding' in headerValues and not headerValues.get('transfer-encoding', 'none') == 'identity':
            if not headerValues['transfer-encoding'] == 'chunked':
                #unsupported transfer encoding
                raise HttpRequestParserException('Unsupported transfer encoding: "%s"', headerValues['transferEncoding'])
            else:
                #chunked encoding
                self.step = 'chunk header'
                self.data = deque()
                self.transferEncoding = 'chunked'
                
                self.log.debug('Conn %5i: Got request (method %s) from client: detected chunked transfer encoding', self.connId, self.header['method'])
                
        elif 'content-length' in headerValues:
            #no transfer encoding
            try:
                length = int(headerValues['content-length'])
            except ValueError:
                raise HttpRequestParserException('Invalid value in "content-length" field: "%s"', headerValues['content-length'])
            
            self.log.debug('Conn %5i: Got request (method %s) from client: detected plain transfer type, expecting "%i" bytes', self.connId, self.header['method'], length)
            
            #adapt state
            self.step = 'body'
            self.data = deque()
            self.transferEncoding = None
            self._increaseDataSizeLimit(length)
            
        else:
            #no transfer at all
            self.log.debug('Conn %5i: Got request (method %s) from client: detected plain header type, no data', self.connId, self.header['method'])
            self.finished = True
    
    
    def _processHeaderData(self, data):
        data = self._getLastChunk() + data
                
        #try to find the endline of the header
        if '\n\n' in data:
            self.newline = '\n'
        elif '\r\n\r\n' in data:
            self.newline = '\r\n'
            
        if self.newline is None:
            #still not finished
            self.log.debug('Conn %5i: Got "%i" bytes of header', self.connId, len(data))
            self._storeChunk(data)
            data = ''
            
        else:
            #found end of header, split data up
            data = data.split(self.newline*2)
            self._storeChunk(data[0])
            data = (self.newline*2).join(data[1:])
            
            #parse header
            self._parseHeader(self._getAllChunks())
            
        return data
    
    
    ##internal functions - body
    
    def _processBodyData(self, data):
        self.log.debug('Conn %5i: Got "%i" bytes of data', self.connId, len(data))
        self._storeData(data)
        if self.dataSize == self.maxDataSize:
            self.finished = True
        elif self.dataSize > self.maxDataSize:
            raise HttpRequestParserException('Received more data then expected: got %i but expected %i!', self.dataSize, self.maxDataSize)
        return ''
    
    
    ##internal functions - chunks
    
    def _processChunkHeaderData(self, data):
        data = self._getLastChunk() + data
        
        if not self.newline in data:
            #line is not finished
            self.log.debug('Conn %5i: Got "%i" bytes of chunk header', self.connId, len(data))
            self._storeChunk(data)
            data = ''
        
        else:
            #finished chunk header
            data = data.split(self.newline)
            self._storeChunk(data[0])
            data = self.newline.join(data[1:])
            
            #decode chunk length
            chunkHeader = self._getAllChunks()
            if ';' in chunkHeader:
                #remove chunk extensions
                chunkHeader = chunkHeader.split(';')[0]
            
            try:
                length = hexToInt(chunkHeader)
            except ValueError:
                raise HttpRequestParserException('Invalid chunk length "%s"', chunkHeader)
            
            if length == 0:
                #got all data
                self.log.debug('Conn %5i: Got final chunk header, waiting for trailer', self.connId)
                self.step = 'trailer'
                
            else:
                #next chunk
                self.log.debug('Conn %5i: Got chunk header, expecting chunk body with "%i" bytes', self.connId, length)
                self.step = 'chunk body'
                self._increaseDataSizeLimit(length)
                self._setMaxBufferSize(length)
        return data
    
    
    def _processChunkBodyData(self, data):
        neededBytes = self.maxBufferSize - self.bufferSize
        newlineLength = len(self.newline)
        
        if len(data) < neededBytes + newlineLength:
            #only a part of the data
            self.log.debug('Conn %5i: Received "%i" bytes of chunk body', self.connId, len(data))
            self._storeChunk(data)
            data = ''
        else:
            #got enough data
            self._storeChunk(data[:neededBytes])
            if not data[neededBytes:neededBytes+newlineLength] == self.newline:
                #chunk is too long
                raise HttpRequestParserException('Received data exceeds expected chunk size')
            else:
                #all ok
                self._storeData(self._getAllChunks())
                self._setMaxBufferSize(self.userHeaderSizeLimit)
                self.log.debug('Conn %5i: Completed chunk body, waiting for next chunk header', self.connId)
                data = data[neededBytes+newlineLength:]
                self.step = 'chunk header'
                
        return data
    
    
    ##internal functions - trailer
    
    def _processTrailerData(self, data):
        data = self._getLastChunk() + data
        if self.bufferSize == 0 and data == self.newline:
            #there is no trailing data at all
            data = self.newline*2
            
        if not self.newline*2 in data:
            #still not finished
            self._storeChunk(data)
            data = ''
            
        else:
            #found end of trailer, split data up
            data = data.split(self.newline*2)
            self._storeChunk(data[0])
            data = (self.newline*2).join(data[1:])
            
            if not data == '':
                raise HttpRequestParserException('Received trailing data after end of response')
            else:
                #parse trailer
                self.log.debug('Conn %5i: Got trailer', self.connId)
                trailerValues = self._parseHeaderValues(self._getAllChunks(), 'trailer')
                headerValues = self.header['values']
                for key, value in trailerValues:
                    if key in headerValues:
                        raise HttpRequestParserException('Received key "%s" both in the header and in the trailer', key)
                    else:
                        headerValues[key] = value
                        
                self.finished = True
        return data
    
    
    ##external functions
        
    def handleData(self, data):
        while len(data) > 0:
            #go on until no data is left or a failure occured
            if self.step == 'header':
                #still need header
                data = self._processHeaderData(data)
                
            elif self.step == 'body':
                #plain data
                data = self._processBodyData(data)
                
            elif self.step == 'chunk header':
                #chunk header
                data = self._processChunkHeaderData(data)
                
            elif self.step == 'chunk body':
                #chunk body
                data = self._processChunkBodyData(data)
                
            elif self.step == 'trailer':
                #trailer
                data = self._processTrailerData(data)
                
        return self.finished
    
    
    def getHeader(self):
        return self.header
    
    
    def getData(self):
        return self._getData()
    
    
    def getAll(self):
        request = {'header':self.header,
                   'data':self._getData()}
        return request