"""
Copyright 2009  Blub

HttpServer, a class which implements a very simple http server. Only a very
small part of the http spec is implemented.
This file is part of PebCache.

PebCache is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation, version 2 of the License.

PebCache is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PebCache.  If not, see <http://www.gnu.org/licenses/>.
"""

##own
import logging
import os

##builtin
from Config import SectionConfig
from FileServer.FileServerHandler import FileServerHandler
from Gcache.GcacheHandler import GcacheHandler
from Http import HttpUtilities
from Http.HttpRequestParser import HttpRequestParser, HttpRequestParserException
from Http.HttpUtilities import HttpUrlParseException
from PySamLib.I2PSocket import I2PSocket
from PySamLib.I2PSocketManager import I2PSocketManager
from Utilities import FileHandle, PseudoFile, logTraceback, quoteUnicode, quoteUnicodePlus, unquoteUnicode, unquoteUnicodePlus




class HttpServer:
    def __init__(self, path, config, version):
        self.path = path
        self.config = config
        self.version = version
        self.log = logging.getLogger('HttpServer')
        
        #get max data sizes
        self.maxHeaderSize = self.config.get('http','maxHeaderSize')
        self.maxBodySize = self.config.get('http','maxBodySize')
        
        #create connection to sam bridge
        self.sam = I2PSocketManager(log=logging.getLogger('Sam'))
        samOptions = {'inbound.nickname':self.config.get('i2p','SamSessionName'),
                      'inbound.quantity':self.config.get('i2p','SamInboundQuantity'),
                      'inbound.backupQuantity':self.config.get('i2p','SamInboundBackupQuantity'),
                      'inbound.length':self.config.get('i2p','SamInboundLength'),
                      'inbound.lengthVariance':self.config.get('i2p','SamInboundLengthVariance'),
                      'inbound.allowZeroHop':self.config.get('i2p','SamInboundAllowZeroHop'),
                      'outbound.quantity':self.config.get('i2p','SamOutboundQuantity'),
                      'outbound.backupQuantity':self.config.get('i2p','SamOutboundBackupQuantity'),
                      'outbound.length':self.config.get('i2p','SamOutboundLength'),
                      'outbound.lengthVariance':self.config.get('i2p','SamOutboundLengthVariance'),
                      'outbound.allowZeroHop':self.config.get('i2p','SamOutboundAllowZeroHop')}

        self.ownDestNum = self.sam.addDestination(self.config.get('i2p', 'SamIp'),
                                                  self.config.get('i2p', 'SamPort'),
                                                  self.config.get('i2p','SamSessionName'),
                                                  'tcp', 'both', samOptions)
        self.ownDest = self.sam.getOwnDestination(destId=self.ownDestNum, timeout=60)  
        self.log.warn('Using Destination "%s"', self.ownDest)

        #connections
        self.allConns = set()
        self.connsWithSendInterest = set()
        self.connsWithRecvInterest = set()
        
        self.conns = {}
        
        #listenConn
        self.listenConn = self.sam.listen(self.ownDestNum, addOld=True)
        self.allConns.add(self.listenConn)
        self.connsWithRecvInterest.add(self.listenConn)
        
        #load handlers
        self.handler = self._loadHandler()
        
        
    def _loadHandler(self):
        self.log.info('Loading handler config...')
        handler = {}
        count = 1
        while self.config.hasOption('Handler','Handler-'+str(count)):
            handlerSection = self.config.get('Handler','Handler-'+str(count))
            if not self.config.hasSection(handlerSection):
                self.log.warn('The config section "%s" for Handler-%d doesn\'t exist!', handlerSection, count)
            else:
                #section exists, load relevant options
                if not (self.config.hasOption(handlerSection, 'HttpUrl') and \
                        self.config.hasOption(handlerSection, 'HttpMethods') and \
                        self.config.hasOption(handlerSection, 'HttpHandlerClass')):
                            self.log.warn('The config section "%s" for Handler-%d misses one of the mandatory Http* options!',
                                          handlerSection, count)
                else:
                    #everything available, so get it
                    handlerUrl = self.config.get(handlerSection, 'HttpUrl')
                    supportedMethods = self.config.get(handlerSection, 'HttpMethods').split(',')
                    handlerClass = self.config.get(handlerSection, 'HttpHandlerClass')
                    self.log.info('Found handler definition for URL "%s" supporting "%s" and using the "%s" class',
                                  handlerUrl, ','.join(supportedMethods), handlerClass)
                                
                    if handlerClass=='Gcache':
                        handlerClass = GcacheHandler(handlerSection, SectionConfig(self.config, handlerSection), self.ownDest, self.path, self.version)
                    elif handlerClass=='FileServer':
                        handlerClass = FileServerHandler(handlerSection, SectionConfig(self.config, handlerSection), self.ownDest, self.path)
                    else:
                        self.log.warn('Handler class "%s" is unknown, ignoring definition!', handlerClass)
                        handlerClass = None
                        
                    if handlerClass is not None:
                        handler[handlerUrl] = {'supportedMethods':supportedMethods,
                                               'handlerClass':handlerClass}
            count += 1
        self.log.info('Finished handler setup')
        return handler
    
    
    def _getHandler(self, request):
        baseUrl = request['header']['url']['path']
        if baseUrl == '':
            baseUrl = '/'
        method = request['header']['method']
        result = None
        for handlerUrl, handlerItems in self.handler.iteritems():
            if ((handlerUrl[-1]=='*' and len(baseUrl)>=(len(handlerUrl)-1) and baseUrl[:len(handlerUrl)-1]==handlerUrl[:-1]) or
                (handlerUrl[-1]!='*' and handlerUrl==baseUrl)):
                #url match
                if method in handlerItems['supportedMethods']:
                    #method is supported
                    result = handlerItems['handlerClass']
                    break;
        return result
    

    def _closeConn(self, connId, force=False):
        connSet = self.conns[connId]
        del self.conns[connId]
        
        #close any open file in the outqueue
        for ele in connSet['outQueue']:
            try:
                ele['file'].close()
            except:
                pass
            
        #cleanup status sets
        self.connsWithSendInterest.discard(connId)
        self.connsWithRecvInterest.discard(connId)
        self.allConns.remove(connId)
        
        #really close conn
        connSet['conn'].close(force)
        
        
    def _buildHeader(self, response):
        header = []
        header.append(response['codeString'])
        header.append('Connection: close')
        header.append('Content-Type: '+response['type'])
        header.append('Content-Length: '+str(response['dataSize']))
        header.append('Cache-Control: no-cache')
        header.append('Accept-Ranges: none')
        header.append('')
        header.append('')
        header = '\r\n'.join(header)
        return header
    
    
    def _sendResponse(self, connId, connSet, request, response):     
        #mark conn for sending
        self.connsWithSendInterest.add(connId)
        
        #check response code
        if response['code']==400:
            #Standard-response: Bad Request
            response['codeString'] = 'HTTP/1.1 400 Bad Request'
            response['type'] = 'text/html'
            data = '<html><head>\n<title>400 Bad Request</title>\n</head><body>\n<h1>Bad Request</h1>\n'\
                   '<p>Your browser sent a request that this server could not understand.</p>\n</body></html>'
            response['data'] = PseudoFile(data)
            response['dataSize'] = response['data'].size()
            
        elif response['code']==403:
            #Standard-response: Forbidden
            response['codeString'] = 'HTTP/1.1 403 Forbidden'
            response['type'] = 'text/html'
            data = '<html><head>\n<title>403 Forbidden</title>\n</head><body>\n<h1>Forbidden</h1>\n'\
                   '<p>You may not access the requested URL %s.</p>\n</body></html>' % HttpUtilities.joinRelativeUrl(request['header']['url'])
            response['data'] = PseudoFile(data)
            response['dataSize'] = response['data'].size()
            
        elif response['code']==404:
            #Standard-response: File not found
            response['codeString'] = 'HTTP/1.1 404 Not Found'
            response['type'] = 'text/html'
            data = '<html><head>\n<title>404 Not Found</title>\n</head><body>\n<h1>Not Found</h1>\n'\
                   '<p>The requested URL %s was not found on this server.</p>\n</body></html>' % HttpUtilities.joinRelativeUrl(request['header']['url'])
            response['data'] = PseudoFile(data)
            response['dataSize'] = response['data'].size()
            
        elif response['code']==500:
            #Standard-response: Internal Server Error
            response['codeString'] = 'HTTP/1.1 500 Internal Server Error'
            response['type'] = 'text/html'
            data = '<html><head>\n<title>500 Internal Server Error</title>\n</head><body>\n<h1>Internal Server Error</h1>\n'\
                   '<p>An internal error occured, please try again later.</p>\n</body></html>'
            response['data'] = PseudoFile(data)
            response['dataSize'] = response['data'].size()
            
        elif response['code']==501:
            #Standard-response: Not Implemented
            response['codeString'] = 'HTTP/1.1 501 Not Implemented'
            response['type'] = 'text/html'
            data = '<html><head>\n<title>501 Not Implemented</title>\n</head><body>\n<h1>Not Implemented</h1>\n'\
                   '<p>The requested http method is not implemented.</p>\n</body></html>'
            response['data'] = PseudoFile(data)
            response['dataSize'] = response['data'].size()
            
        elif response['code']==200:
            #normal response, send data back
            response['codeString'] = 'HTTP/1.1 200 Ok'
            
            #get data
            if 'file' in response:
                #send a file as a response
                if isinstance(response['file'], basestring):
                    response['data'] = FileHandle(response['file'])
                else:
                    response['data'] = response['file']
            else:
                #simple string
                strData = response['string']
                if isinstance(strData, unicode):
                    strData = strData.encode('UTF-8', 'ignore')
                response['data'] = PseudoFile(strData)
                
            #get data size
            try:
                response['dataSize'] = response['data'].size()
            except:
                #send error response
                self.log.warn('Conn %5i: Failed to get the size of the response to url "%s".', connId, HttpUtilities.joinRelativeUrl(request['header']['url']))
                response['codeString'] = 'HTTP/1.1 500 Internal Server Error'
                response['type'] = 'text/html'
                data = '<html><head>\n<title>500 Internal Server Error</title>\n</head><body>\n<h1>Internal Server Error</h1>\n'\
                       '<p>An internal error occured, please try again later.</p>\n</body></html>'
                response['data'] = PseudoFile(data)
                response['dataSize'] = response['data'].size()
        
        else:
            #Standard-response: Internal Server Error
            self.log.warn('Conn %5i: Invalid response code "%s" for response to "%s"!', connId, str(response['code']), HttpUtilities.joinRelativeUrl(request['header']['url']))
            response['codeString'] = 'HTTP/1.1 500 Internal Server Error'
            response['type'] = 'text/html'
            data = '<html><head>\n<title>500 Internal Server Error</title>\n</head><body>\n<h1>Internal Server Error</h1>\n'\
                   '<p>An internal error occured, please try again later.</p>\n</body></html>'
            response['data'] = PseudoFile(data)
            response['dataSize'] = response['data'].size()
        
        response['dataWritten'] = 0
        
        #build header
        header = self._buildHeader(response)
        
        #queue header
        queueEle = {'file':PseudoFile(header),
                    'size':len(header),
                    'send':0}
        connSet['outQueue'].append(queueEle)
        
        #queue data
        if request['header']['method'] == 'HEAD':
            #strip the real response data for HEAD requests
            response['data'].close()
            
        elif response['dataSize'] > 0:
            #add tp queue
            queueEle = {'file':response['data'],
                        'size':response['dataSize'],
                        'send':0}
            connSet['outQueue'].append(queueEle)
        
        
    def _sendDefaultResponse(self, connId, connSet, request, responseCode):
        if request is None:
            #dummy
            request = {'data':'',
                       'header':{'method':'',
                                 'url':{'address':'', 'path':''},
                                 'version':''}}
        response = {'code': responseCode}
        self._sendResponse(connId, connSet, request, response)
        
        
    ##socket accept/read/send
    
    def _acceptNewConns(self):
        newConns = self.sam.accept(self.listenConn)
        for newConn in newConns:
            self.log.info('Accepted new connection from "%s" with ID "%d"', newConn[1][:10], newConn[0])
            self.allConns.add(newConn[0])
            self.connsWithRecvInterest.add(newConn[0])
            self.conns[newConn[0]] = {'conn':I2PSocket(self.sam, self.ownDestNum,  newConn[0], newConn[1], 'tcp'),
                                      'requestParser':HttpRequestParser(newConn[0], self.maxHeaderSize, self.maxBodySize),
                                      'outQueue':[]}
            
    
    def _connRecvEvent(self, connId):
        #data from connection
        connSet = self.conns[connId]
        newData = connSet['conn'].recv()
        self.log.debug('Conn %5i: Got %d bytes of data.', connId, len(newData))
        
        #parse data
        try:
            finished = connSet['requestParser'].handleData(newData)
            if not finished:
                self.log.debug('Conn %5i: Got partial request, waiting for more data.', connId)
            else:
                #finished
                self.connsWithRecvInterest.remove(connId)
                request = connSet['requestParser'].getAll()
                header = request['header']
                self.log.debug('Conn %5i: Got valid request: "%s %s %s"', connId, header['method'], HttpUtilities.joinUrl(header['url']), header['version'])
                if header['method'] not in ('GET', 'HEAD'):
                    #not supported
                    self._sendDefaultResponse(connId, connSet, request, 501)
                else:
                    #ok
                    handler = self._getHandler(request)
                    if handler is None:
                        #no matching handler
                        self.log.info('Conn %5i: No handler for url "%s"', connId, HttpUtilities.joinRelativeUrl(header['url']))
                        self._sendDefaultResponse(connId, connSet, request, 404)
                    else:
                        #let handler process request
                        try:
                            response = handler.doGet(connId, connSet['conn'].getpeername(), request)                                       
                        except:
                            self.log.warn('Conn %5i: Handler for url "%s" crashed!\nTraceback:\n%s', connId, HttpUtilities.joinRelativeUrl(header['url']), logTraceback())
                            response = {'code':500}
                        self._sendResponse(connId, connSet, request, response)
            
        except HttpRequestParserException, e:
            self.connsWithRecvInterest.discard(connId)
            self.log.warn('Conn %5i: Failed to parse request: %s', connId, e.reason)
            self._sendDefaultResponse(connId, connSet, None, 400)
            
            
        except:
            self.connsWithRecvInterest.discard(connId)
            self.log.warn('Conn %5i: Internal error, traceback:\n%s', connId, logTraceback())
            self._sendDefaultResponse(connId, connSet, None, 500)
    
    
    def _connSendEvent(self, connId):
        connSet = self.conns[connId]
        assert connId in self.connsWithSendInterest, 'trying to send but not interested? ID: \"%d\"' % connId
        assert len(connSet['outQueue']) > 0, 'trying to send but empty queue?! ID: \"%d\"' % connId
        
        #get free buffer space (safe because only this thread does anything with this conn id)
        freeBufferSpace = connSet['conn'].getFreeOutBufferSpace()
        
        #check what needs to be read
        dataSet = connSet['outQueue'][0]
        wantedBytes = min(freeBufferSpace, (dataSet['size'] - dataSet['send']))
        assert wantedBytes > 0, 'nothing to send?!'
        
        #read
        try:
            data = dataSet['file'].read(wantedBytes)
            dataLen = len(data)
            success = True
        except:
            success = False
        
        if not success:
            #file operation failed completly
            self.log.warn('Conn %5i: File operation failed, closing conn!', connId)
            self._closeConn(connId, force=True)
        
        elif dataLen == 0:
            #file operation returned nothing, that counts as a failure
            self.log.warn('Conn %5i: Read return 0 bytes, file size is smaller then expected (expected %i, got %i): closing conn', connId, dataSet['size'], dataSet['send']+dataLen)
            self._closeConn(connId, force=True)
            
        else:
            #at least got some data
            if dataLen < wantedBytes:
                self.log.warn('Conn %5i: Abnormal file condition, read returned less bytes then told!', connId)
                
            #adjust counter
            dataSet['send'] += dataLen
            assert dataSet['send'] <= dataSet['size'], 'sent more then should exist?!'
            
            #send data, but don't check return: only case when not all data gets send is if the conn failed
            #in the meantime, in this case we will know that after the next select call, no harm done
            self.log.debug('Conn %5i: Sending %d bytes of data.', connId, len(data))
            connSet['conn'].send(data) 
            
            #check if finished
            if dataSet['send'] == dataSet['size']:
                #finished with this file
                self.log.debug('Conn %5i: Finished with one response file.', connId)
                try:
                    dataSet['file'].close()
                except:
                    pass
                del connSet['outQueue'][0]
                if len(connSet['outQueue']) == 0:
                    #response send, close conn
                    self.log.debug('Conn %5i: Finished sending response, closing.', connId)
                    self._closeConn(connId)
                    
                    
    def run(self):
        try:
            stopfile = os.path.join(self.path,'stop')
            while not os.path.isfile(stopfile):
                #select
                readable, writeable, errored = self.sam.select(self.connsWithRecvInterest, self.connsWithSendInterest, self.allConns, timeout=1.0)
                
                for connId in errored:
                    assert connId != self.listenConn,'Listen socket failed?!'
                    self.log.info('Conn %5i: conn failed, closing', connId)
                    self._closeConn(connId)
                        
                for connId in readable:
                    if not connId in self.allConns:
                        self.log.warn('Conn %5i: readable but not known to us?!', connId)
                    else:
                        assert connId in self.connsWithRecvInterest, 'trying to receive but not interested? ID: \"%d\"' % connId
                        if connId == self.listenConn:
                            self._acceptNewConns()
                        else:
                            self._connRecvEvent(connId)
                            
                for connId in writeable:
                    if connId in self.allConns:
                        self._connSendEvent(connId)
            
            self.log.warn('Terminating...')
            
            #close all conns
            self.log.warn('Closing alls open connections')
            for connId in list(self.allConns):
                if not connId == self.listenConn:
                    self._closeConn(connId, force=True)
            
            #close listening socket
            self.log.warn('Closing listening socket')
            self.allConns.remove(self.listenConn)
            self.connsWithRecvInterest.remove(self.listenConn)
            self.sam.close(self.listenConn, force=True)
            
        except:
            self.log.warn('Main loop crashed!\nTraceback:\n%s', logTraceback())
                
        #close connection to SAM
        self.log.warn('Removing sam destination')
        self.sam.removeDestination(self.ownDestNum)
        
        self.log.warn('Terminating thread')